unit uDatosHistoricos;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
{$IFDEF FPC}
//  FileUtil,
{$ENDIF}
  uFechas, Classes, xMatDefs, SysUtils, uAuxiliares, DateUtils;


type
  TDatosHistoricos = class
    private
      series: TMatOfNReal;
      nPuntosDelAnillo: integer; // cantidad de puntos del anillo
      procedure calcNAniosDatos;
    public

      //Cantidad de anios de datos del archivo
      nAniosDatos_Min: Integer;
      nAniosDatos_Max: integer;

      fechaIni, fechaFin: TFecha;
      nSeries, nPuntosSeries, nPuntosPorCiclo: Integer;
      NombresDeBornes_Publicados: TStringList;
      dt_EntrePuntos, inv_dt_EntrePuntos: TDateTime;
      anioIni, mesIni, diaIni: word;

      constructor Create(fechaIni: TFecha;
                         NombresDeBornes_Publicados: TStringlist;
                         series: TMatOfNReal;
                         nPuntosPorCiclo: Integer;
                         dt_EntrePuntosdt: TDateTime);

      // carga los datos de un archivo previamente guardado con WriteToArchi
      constructor CreateFromArchi(nombreArchivo: String);

      // escribe los datos en un archivo de texto.
      procedure WriteToArchi( nombreArchivo: string );
      procedure Free;

      procedure CalcularConstantes;

      // retorna el ndice dentro de los datos histricos de forma tal
      // que la fecha del dato al que corresponde el ndice sea <= que la fecha
      // pasada como parametro. Si kAnio es >= 0 es un "desplazamiento" dentro
      // del anillo de datos. A la fecha pasada como
      // parmetro primero se le cambia el ao por el ao de inicio de los datos
      // ms kAnio y luego se busca, si la fecha as resultante est fuera de
      // la ventada de datos se le resta la cantidad de aos de datos tantas veces
      // como sea necesario para encontrar una fecha dentro de la ventana de datos.
      // esto es equivalente a suponer que los datos histricos forman un anillo
      // y que cuando nos pasamos del final comienza por el principio, pero siempre
      // teniendo cuidado de sincronizar dentro del ao.
      function locate_fecha_ignore_anio_(fecha: TFecha; kAnio: integer;
        var dt_desp: NReal): integer;

      // Busca la fecha pasada como parmetro
      // esto es usado para reproducir la cronica histrica sincronizada
      // con el tiempo de las salas de simulacin. El resultado < 0 implica
      // que no se encuentra la fecha buscada en la ventana de datos.
      function locate_fecha_( fecha: TFecha;
        var dt_desp: NReal ): integer;


      // fija la dimensin de los arrays de indices y pesos de interpolacin
      // para un paso de simulacin den ancho dt_paso.
      procedure setLength_indices_y_pesos(var kIndices: TDAOfNInt;
        var pesos: TDAOfNreal; dt_Paso: NReal);

      // por eficiencia se supone que kIndices y Pesos ya estn creados
      // de la dimensiones que corresponde.
      // Para ello se puede usar SetLength_indices_y_pesos
      procedure calc_indices_y_pesos(const AniosDesp: integer; const fecha: TFecha;
        const dt_Paso: NReal; var kIndices: TDAOfNint; var pesos: TDAOfNReal); // pesos para interpolar

      // da el valor medio de un paso de tiempo que comienza
      // en el indice dado, con un desfasaje dt_desp  de ancho dt_paso
      function get_mval_( iSerie: integer; const kIndices: TDAOfNInt; const pesos: TDAOfNReal ): NReal;
  end;

implementation


procedure TDatosHistoricos.CalcularConstantes;
begin
  inv_dt_EntrePuntos:= 1 / dt_EntrePuntos;

  fechaFin:= FechaIni.Create_Clone( fechaIni );
  fechaFin.addDias( trunc( self.nPuntosSeries * self.dt_EntrePuntos +0.1 ));
  calcNAniosDatos;
  nPuntosDelAnillo:= (nPuntosSeries div nPuntosPorCiclo) * NPuntosPorCiclo;
end;


function TDatosHistoricos.locate_fecha_ignore_anio_( fecha: TFecha; kAnio: integer; var dt_desp: NReal ): integer;
var
  res: integer;
  ft: TFecha;
  cnt: integer;
  buscando: boolean;

begin
  kAnio:= kAnio mod ( self.nAniosDatos_Max -1 );
  cnt:= 0;
  res:= -1;
  buscando:= true;
  ft:= TFecha.Create_Clone( fecha );
  while ( buscando and ( cnt <= 3 ) ) do
  begin
    ft.SetAnio( self.anioIni + kAnio );
    res:= locate_fecha_( ft, dt_desp );
    if res >= 0 then
      buscando:= false
    else
    begin
     inc( cnt );
     if kAnio = 0 then
      inc( kAnio )
    else
      kAnio:= 0;
    end;
  end;
  result:= res;
end;

function TDatosHistoricos.locate_fecha_( fecha: TFecha; var dt_desp: NReal ): integer;
var
  rdt: double;
begin
  if (fecha.dt < fechaIni.dt ) or ( fecha.dt >= fechaFin.dt ) then
  begin
    result:= -1;
  end
  else
  begin
    rdt:= (fecha.dt + medio_Segundo - fechaIni.dt )* inv_dt_EntrePuntos;
    dt_desp:= frac( rdt );
    result:= trunc( rdt);
  end;
end;


// fija la dimensin de los arrays de indices y pesos de interpolacin
// para un paso de simulacin den ancho dt_paso.
procedure TDatosHistoricos.setLength_indices_y_pesos(
                   var kIndices: TDAOfNInt;
                   var pesos: TDAOfNreal;
                   dt_Paso: NReal );
var
  n: integer;
begin
  n:= trunc( ( dt_Paso - medio_Segundo  ) / dt_EntrePuntos ) + 2;
  setlength( kIndices, n );
  setlength( pesos, n );
end;

procedure TDatosHistoricos.calc_indices_y_pesos(
            const AniosDesp: integer; // anios de desplazamiento
            const fecha: TFecha; // fecha de inicio del paso
            const dt_Paso: NReal; // ancho del paso de simulacin
            var kIndices: TDAOfNint; // indices para interpolar
            var pesos: TDAOfNReal ); // pesos para interpolar

var
  k, j: integer;
  desp: NReal;
  p: NReal;
  dt_base: NReal;
  rk: NReal;
begin
  dt_base:= self.fechaIni.dt + AniosDesp * DiasDelAnioMedio;
  rk:= (fecha.dt - dt_base)/dt_EntrePuntos;
  desp:= frac( rk );
  k:= trunc( rk ) mod nPuntosDelAnillo;
//  k:= locate_fecha_ignore_anio( fecha, AniosDesp, desp );
  kIndices[0]:= k;
  p:= ( dt_EntrePuntos* (1 - desp ))/ dt_Paso;
  pesos[0]:= p;

  for j:= 1 to high( pesos )-1 do
  begin
    inc( k );
    if k >= self.nPuntosSeries then
     k := 0;
    kIndices[j]:= k;
    pesos[j]:= dt_EntrePuntos/ dt_Paso;
  end;

  j:= high( pesos );
  desp:= dt_paso -( ( high( pesos ) -  desp )* dt_EntrePuntos   );
  if desp > 0 then
  begin
   inc( k );
   if k >= self.nPuntosSeries then
     k := 0;
  end;
  kIndices[ j ]:= k;
  pesos[ j ]:=  desp / dt_paso;
end;


function TDatosHistoricos.get_mval_( iSerie: integer; const kIndices: TDAOfNInt; const pesos: TDAOfNReal ): NReal;
var
  k: integer;
  res: NReal;
  aSerie: TDAOfNReal;
begin
  aSerie:= series[iSerie];
  res:= aSerie[ kIndices[0] ] * pesos[0];
  for k:= 1 to high( pesos ) do
    res:= res + aSerie[ kIndices[k] ] * pesos[k];
  result:= res;
end;

constructor TDatosHistoricos.Create(fechaIni: TFecha;
                                    NombresDeBornes_Publicados: TStringlist;
                                    series: TMatOfNReal;
                                    nPuntosPorCiclo: Integer;
                                    dt_EntrePuntosdt: TDateTime);
begin
  inherited Create;
  self.fechaIni:= fechaIni;
  self.NombresDeBornes_Publicados:= NombresDeBornes_Publicados;
  self.series:= series;
  self.nPuntosPorCiclo:= nPuntosPorCiclo;
  self.nSeries:= Length(series);
  self.nPuntosSeries:= Length(series[0]);
  self.dt_EntrePuntos:= dt_EntrePuntosdt;
  DecodeDate(fechaIni.dt, anioIni, mesIni, diaIni);
  CalcularConstantes;
end;

constructor TDatosHistoricos.CreateFromArchi(nombreArchivo: String);
var
  f: TextFile;
  i, j: Integer;
  linea: String;
  hora, minuto, segundo: word;
  dtFechaIni: TDateTime;
  horasEntrePuntos: NReal;
  old_filemode: integer;
begin
  if FileExists(nombreArchivo) then
  begin
    AssignFile(f, nombreArchivo);

    old_FileMode:= filemode;
    filemode:= 0;
    Reset(f);
    filemode:= old_FIlemode;
    uauxiliares.setSeparadoresGlobales;
    try
      //nSeries
      readln(f, linea);
      eliminar_BOM( linea );
      nSeries:= NextInt(linea);

      //fechaIni
      Readln(f, linea);
      anioIni:= NextInt(linea);
      mesIni:= NextInt(linea);
      diaIni:= NextInt(linea);
      hora:= NextInt(linea);
      minuto:= NextInt(linea);
      segundo:= NextInt(linea);
      dtFechaIni:= EncodeDateTime(anioIni, mesIni, diaIni, hora, minuto, segundo, 0);
      fechaIni:= TFecha.Create_Dt(dtFechaIni);

      //Perodo de muestreo
      Readln(f, linea);

      horasEntrePuntos:= NextFloat(linea);

      // Atencin, si pusieron 168.5xx lo imagino que se trata de un archivo
      // con datos arreglados a 52 valores por ao y pongo el valor exacto
      // de la duracin para que no se desincronizen los aos en simulaciones
      // muy largas.
      if abs( horasEntrePuntos - HorasSemana52 ) < 0.25 then
         horasEntrePuntos:= HorasSemana52;

      dt_EntrePuntos:=  horasEntrePuntos * horaToDt;
      inv_dt_EntrePuntos:= 1.0 / dt_EntrePuntos;

      //nPuntos
      readln(f, linea);
      nPuntosSeries:= NextInt(linea);
      calcNAniosDatos;

      //NPuntos por Ciclo
      readln(f, linea);
      nPuntosPorCiclo:= NextInt(linea);

      SetLength(series, nSeries);
      for i:= 0 to nSeries - 1 do
        SetLength(series[i], nPuntosSeries);

      //Encabezados Series
      self.NombresDeBornes_Publicados:= TStringList.Create;
      readln(f, linea);
      for i:= 0 to nSeries - 1 do
        NombresDeBornes_Publicados.Add(NextPal(linea));

      for j:= 0 to nPuntosSeries - 1 do
      begin
        readln(f, linea);
        NextPal(linea);
        for i:= 0 to nSeries - 1 do
          series[i][j]:= NextFloat(linea);
      end;

      CalcularConstantes;

    finally
      uauxiliares.setSeparadoresLocales;
      CloseFile(f);
    end;
  end
  else
    Raise Exception.Create('TDatosFuenteHistorica.CreateFromArchi: no se encuentra el archivo ' + nombreArchivo);
end;

procedure TDatosHistoricos.WriteToArchi( nombreArchivo: string );
var
  f: TextFile;
  i, j: Integer;
  linea: String;
  anio, mes, dia, hora, minuto, segundo, ms: word;
begin
  AssignFile(f, nombreArchivo);
  try
    Rewrite(f);

    writeln(f, nSeries, #9'NSeries');
    DecodeDateTime(fechaIni.dt, anio, mes, dia, hora, minuto, segundo, ms);
    writeln(f, anio, #9, mes, #9, dia, #9, hora, #9, minuto, #9, segundo, #9'FechaIni');
    writeln(f, dt_EntrePuntos * dtToHora, #9'Perodo de muestreo[h]');
    writeln(f, nPuntosSeries, #9'NPuntosSeries');
    writeln(f, nPuntosPorCiclo, #9'Puntos por ciclo');

    linea:= '';
    for i:= 0 to NombresDeBornes_Publicados.Count - 1 do
      linea:= linea + #9 + NombresDeBornes_Publicados[i];
    Writeln(f, linea);

    for j:= 0 to nPuntosSeries - 1 do
    begin
      linea:= IntToStr(j) + #9 + FloatToStr(series[0][j]);
      for i:= 1 to nSeries - 1 do
        linea:= linea + #9 + FloatToStr(series[i][j]);
      writeln(linea);
    end;
  finally
    CloseFile(f);
  end;
end;

procedure TDatosHistoricos.Free;
var
  i: Integer;
begin
  fechaIni.Free;
  NombresDeBornes_Publicados.Free;
  for i:= 0 to high(series) do
    SetLength(series[i], 0);
  SetLength(series, 0);
  inherited Free;
end;

procedure TDatosHistoricos.calcNAniosDatos;
(*
var
  anioIni, anioFin, mes, dia: Word;
  dtFin: TDateTime;
  ndias: double;
  *)
begin
(*
  dtFin:= fechaIni.dt + (nPuntosSeries - 1) * tiempoEntrePuntos;
  DecodeDate(fechaIni.dt, anioIni, mes, dia);
  DecodeDate(dtFin, anioFin, mes, dia);
  nAniosDatos:= anioFin - anioIni + 1;
*)
// La cantidad de aos est calculada para un ao de 365 das.
// Los bisiestos se supone que no molestan. La idea es determinar
// la cantidad de "aos completos" que cubre una serie de datos
// si son 365 puntos de 1 da de tiempoEntrePuntos cubre un ao!.
  nAniosDatos_Min:= trunc( nPuntosSeries * dt_EntrePuntos  / 365 +0.5 );
  nAniosDatos_Max:= nAniosDatos_Min+1;
end;

end.
